from rich import print
from rich.console import Console
from rich.table import Table
from rich.panel import Panel
from rich.padding import Padding
from rich.live import Live
from datetime import datetime
from odk_mailer.lib import log
import typer
import time
import base64
import sys

# return unformatted string instead of raising error
# when key is missing within dictionary
# https://stackoverflow.com/a/17215533/3127170
class SafeDict(dict):
    def __missing__(self, key):
        return '{' + key + '}' 

def abort(msg):
    print("[bold red]Error:[/bold red] "+msg)
    raise typer.Abort()

def now():
    return int(time.time())

def ts_to_str(ts):
    dt = datetime.fromtimestamp(ts)
    return dt.strftime("%c")

def base64_encode_str(input_str: str):
    input_bytes = input_str.encode("ascii") 
    base64_bytes = base64.b64encode(input_bytes)

    return base64_bytes.decode("ascii")

def base64_decode_str(base64_str: str):
    base64_bytes = base64_str.encode("ascii") 
    output_bytes = base64.b64decode(base64_bytes)
    
    return output_bytes.decode("ascii")

def safe_format_map(text, dict):
    return text.format_map(SafeDict(vars(dict)))

def render_state(state):
    match state:
        case 0:
            return "pending"
        case 1: 
            return "success"
        case _:
            return "failure"

def print_jobs(jobsMeta):
    console = Console()
    header = ["Job Id", "Created","Scheduled", "Recipients", "State", "Note"]
    table = Table(*header, expand=False, highlight=True, box=None, title_justify="left", show_lines="True")
    console.print()
    for job in jobsMeta:
        row = []
        for key,val in job.items():
            if key == 'hash':
                row.append(val[:10])
            if key in ['created','scheduled']:
                row.append((ts_to_str(val)))
            if key== 'recipients':
                row.append(str(val))
            if key == 'state':
                row.append(render_state(val))
            if key == 'note':
                row.append(val)
        # data = [job["hash"], job["created"], job["scheduled"], job["state"], job["source"], job["title"]]
        # row.extend(data)
        table.add_row(*row)
    console.print(table)
    console.print()

def print_panel(data):
    print(Panel(data))

def print_mail(content):
    print(Padding(content, 1))

def print_success(msg):
        print(f"[bold green]✅ {msg}[/bold green]")


from odk_mailer.classes.Source import Source
from odk_mailer.classes.Fields import Fields
from odk_mailer.classes.Message import Message
from odk_mailer.classes.Schedule import Schedule
from odk_mailer.classes.JobMeta import JobMeta


from odk_mailer.lib import globals, utils,db
import os
import json
import sys
from json import JSONEncoder
import hashlib
from types import SimpleNamespace

# return unformatted string instead of raising error
# when key is missing within dictionary
# https://stackoverflow.com/a/17215533/3127170
class SafeDict(dict):
    def __missing__(self, key):
        return '{' + key + '}' 

# custom encoder for JSONEncoder
# https://pynative.com/python-convert-json-data-into-custom-python-object/
class MailjobEncoder(JSONEncoder):
    def default(self, o):
        return o.__dict__
    
class Mailjob:
    source: Source
    fields: Fields
    message: Message
    schedule: Schedule
    created: int

    hash: str
    json: str

    def __init__(self, hash_or_id = "") -> None:

        if hash_or_id != "":
            self.hash = hash_or_id
            self.hash = self.find()
            self.load()

    def find(self):
        if not self.hash:
            raise Exception("Invalid operation. Hash required")
        
        jobs_meta = db.read_meta()
        found = next((obj for obj in jobs_meta if obj["hash"].startswith(self.hash)), None)

        if not found:
            utils.abort("Job not found.")        

        return found["hash"]

    def load(self):

        with open(os.path.join(globals.odk_mailer_jobs, self.hash+'.json'), 'r', encoding='utf-8') as f:
            job = json.load(f, object_hook=lambda d: SimpleNamespace(**d))
        
        # this may be optimized...
        self.source = Source(job.source)
        self.field = Fields(job.fields)
        self.message = Message(job.message)
        self.schedule = Schedule(job.schedule)
        self.recipients = job.recipients

        self.created = job.created

    def create(
        self, 
        source: Source, 
        fields: Fields, 
        message: Message, 
        schedule: Schedule
        ) -> str:

        self.source = source
        self.fields = fields
        self.message = message
        self.schedule = schedule
        self.created = utils.now()

        self.setRecipients()
        self.json = json.dumps(vars(self), ensure_ascii=True, indent=4, cls=MailjobEncoder)
        self.hash = hashlib.sha256(self.json.encode()).hexdigest()

        self.save()
        return self.hash

    def setRecipients(self):
        recipients = []

        for row in self.source.get_rows():   
            f_row = {k: row[k] for k in self.fields.data + [self.fields.email] }
            recipients.append(f_row)
        self.recipients = recipients

    def save(self)->None:
        # add to jobs
        path_to_job = os.path.join(globals.odk_mailer_jobs, self.hash+'.json')
        with open(path_to_job, 'w', encoding='utf-8') as f:
            f.write(self.json)

        # add to meta
        jobs_meta = db.read_meta()
        jobs_meta.append(JobMeta(self).__dict__)
        db.write_meta(jobs_meta)
    
    #tbd: def delete(self):

    def updateState(self, state):

        if not state in [0,1,2]:
            raise Exception("Invalid state")

        def mapNewMeta(job_meta):
            if job_meta["hash"] == self.hash:
                 job_meta["state"] = state

            return job_meta

        jobs_meta = list(map(mapNewMeta, db.read_meta()))
        db.write_meta(jobs_meta)




        




from odk_mailer.lib import utils, log
from dotenv import load_dotenv
import os

class Config():
    odk_host: str
    smtp_host: str
    smtp_port: str
    smtp_user: str
    smtp_pass: str
    
    def __init__(self, smtpRequired:bool=False, odkRequired:bool=False):
        
        # loads environment variables from .env file
        load_dotenv()

        ODK_HOST = os.getenv('ODK_HOST')

        SMTP_HOST = os.getenv('SMTP_HOST')
        SMPT_PORT = os.getenv('SMTP_PORT')
        SMPT_USER = os.getenv('SMPT_USER')
        SMPT_PASS = os.getenv('SMPT_PASS')

        if smtpRequired:
            if not SMTP_HOST or not SMPT_PORT:
                log.write("Config Error: SMTP_HOST and SMPT_PORT are required configuration for sending emails.", "error")
                utils.abort("Config Error: SMTP_HOST and SMPT_PORT are required configuration for sending emails.")

        if odkRequired and not ODK_HOST:
            log.write("Config Error: ODK_Host is required configuration for accessing ODK API.", "error")
            utils.abort("Config Error: ODK_Host is required configuration for accessing ODK API.")

        self.smtp_host = SMTP_HOST
        self.smtp_port = SMPT_PORT
        self.smtp_user = SMPT_USER
        self.smtp_pass = SMPT_PASS

        self.odk_host = ODK_HOST     

    def get_smpt_config(self):
        return [self.smtp_host, self.smtp_port, self.smtp_user, self.smtp_pass]
        
    def print(self):

        print()
        print("Your current configuration is: ")    
        for key,value in vars(self).items():
            print(f"{key.upper()}: {value}")
        print()
        print("To change your configuration please setup/edit environment variables or .env file!")


from odk_mailer.lib import utils
from types import SimpleNamespace

class Message:
    sender: str
    subject: str
    source: str
    location: str
    format: str
    content: str

    def __init__(self, data) -> None:

        if type(data) is dict:
            self.sender = data["sender"]
            self.subject = data["subject"]
            self.source = data["source"]
            self.format = data["type"]
            self.location = data["location"]            
            self.setContent()

        if type(data) is SimpleNamespace:
            self.sender = data.sender
            self.subject = data.subject
            self.source = data.source
            self.format = data.format
            
            # backwards compatibility
            try:
                self.location = data.location
                self.setContent()
            except:
                self.content = data.content

    def setContent(self):
        if self.source == "stdin":
                self.content = self.location

        if self.source == "path":            
            with open(self.location) as f:
                data_string = f.read()
                # base64_string = utils.base64_encode_str(data_string)
                # self.content = base64_string
                self.content = data_string

        if self.source == "url":
            #tbd: read from url and store as base64
            self.content = self.location


from pathlib import Path
from odk_mailer.lib import globals, cron
import typer
import json

def init():

  # create base dir if not exists
  path_base = Path(globals.odk_mailer_base)
  path_base.mkdir(parents=True, exist_ok=True)
  
  # create jobs dir if not exists
  path_job = Path(globals.odk_mailer_jobs)
  path_job.mkdir(parents=True, exist_ok=True)

  # create jobs.json if not exists
  path_meta = Path(globals.odk_mailer_meta)

  if not path_meta.exists():
      path_meta.touch()

  if path_meta.stat().st_size == 0:
      path_meta.write_text("[]")

  if not is_json(path_meta.read_text()):
      raise typer.Exit(f"The file at '{path_meta}' is invalid.\n") 

def is_json(myjson):
  try:
    json.loads(myjson)
  except ValueError as e:
    return False
  return True


from odk_mailer.lib import log, utils
import smtplib

def send(email, verbose=False, smtp_config=[]):

    if not smtp_config:
        log.write("Fatal Error: Invalid configuration given during smtp.send call.", "error")
        raise Exception("Fatal Error: No Configuration File found.")
    
    smtp_host, smtp_port, smtp_user, smtp_pass = smtp_config       

    try:            
        smtp = smtplib.SMTP(timeout=5)
        if verbose:
        # enable debugging by CLI flag --debug
            smtp.set_debuglevel(2)

        smtp.connect(smtp_host, smtp_port)

        if smtp_user and smtp_pass:
        # if username and password are supplied, smtp.login()            
            smtp.login(smtp_user, smtp_pass)
        
        smtp.send_message(email)
        smtp.quit()

        log.write(f"Successfully sent email to {email['To']}")

        return [ utils.now(), False]
    
    except Exception as error:
        log.write("Failed sending mail to: " + email["To"], error)
        log.write(error, "error")

        return [None, error]


from odk_mailer.classes import Mailjob
from odk_mailer.classes.Config import Config
from odk_mailer.lib import utils, smtp, log
from rich.progress import track
from email.message import EmailMessage
import time
import sys


class Mailer:
    mailjob: Mailjob
    verbose: bool
    dry: bool

    def __init__(self, dry=False, verbose=False) -> None:
        self.dry = dry
        self.verbose = verbose

    def send(self, mailjob: Mailjob):

        self.mailjob = mailjob
        success = []
        errors = []

        if not self.dry:
            log.write(f"Running mailjob {self.mailjob.hash}")

        if self.verbose:           
            print()
            print(f"Running mailjob {self.mailjob.hash}")            
            if self.dry:
                print("Dry Run enabled -no Emails will be sent.")
            print()

        # Decode base64 if needed (currently disabled for issues with ISO decoding of ä characters)
        if self.mailjob.message.source == "stdin":
            text = self.mailjob.message.content
        if self.mailjob.message.source in ["path", "url"]:
            #text = utils.base64_decode_str(self.mailjob.message.content)
            text = self.mailjob.message.content

        # Retrieve SMTP Config with required parameters
        smtp_config = Config(smtpRequired=True).get_smpt_config()
        total_recipients= len(self.mailjob.recipients)

        if self.dry:
            print()
            print("Message Summary")
            print("-----------------------------------------")
            print(f"sender: {self.mailjob.message.sender}")
            print(f"subject: {self.mailjob.message.subject}")
            print(f"source: {self.mailjob.message.source}")
            print(f"format: {self.mailjob.message.format}")
            print(f"# recipients: {total_recipients}")
            print("-----------------------------------------")
            print()

            for idx, recipient in enumerate(self.mailjob.recipients):
                content = utils.safe_format_map(text, recipient)
                print(f"# {idx+1}|{total_recipients} ({recipient.email})")
                print()
                utils.print_mail(content)

        else:
            total = 0

            if self.verbose:
                print(f"Preparing to send {total_recipients} mails")
            log.write(f"Preparing to send {total_recipients} mails")

            for recipient in track(self.mailjob.recipients, description="Processing.."):

                content = utils.safe_format_map(text, recipient)

                email = EmailMessage()
                email['From'] = self.mailjob.message.sender
                email['To'] = recipient.email
                email['Subject'] = self.mailjob.message.subject
                email.set_content(content, subtype=self.mailjob.message.format)

                log.write(f"#{total} of {total_recipients}: Attempting to send mail to {email['To']}")
                
                if self.verbose:
                    print()
                    print(f"#{total} of {total_recipients}: Attempting to send mail to {email['To']}")

                [sent, error] = smtp.send(email, self.verbose, smtp_config)

                if sent:
                    success.append(sent)
                    if self.verbose:
                        print()
                        print("Successfully sent email to " + email["To"])
                        print()                    

                if error:
                    errors.append(error)
                    if self.verbose:
                        print(error)
                        print("Failed sending mail to: " + email['To'])
                        print()                    

                # Fake processing time
                # time.sleep(0.01)
                total += 1
            
            print(f"Processed {total} mails with {len(success)} success and {len(errors)} failure.")            
            log.write(f"Processed {total} mails with {len(success)} success and {len(errors)} failure.")

        return [ len(errors) > 0, errors]


import json
from odk_mailer.lib import globals

def read_meta(hash_or_id: str = ''):

    with open(globals.odk_mailer_meta, "r") as f:
        jobs_meta =json.load(f)

        # return all as list
        if not hash_or_id:
            return jobs_meta
        
        # return single with
        job_meta = next((obj for obj in jobs_meta if obj["hash"].startswith(hash_or_id)), None)
    
        return job_meta

def write_meta(jobs_meta_new):
     
     with open(globals.odk_mailer_meta, "w") as f:
         f.write(json.dumps(jobs_meta_new))
    


import time
from datetime import datetime
from types import SimpleNamespace

class Schedule:
    now: bool   # can be removed in future
    timestamp: int

    def __init__(self, data):
        
        if type(data) is dict:
            self.now = data["now"]        

            if self.now:
                self.timestamp = int(time.time())

            else:
                _datetime = datetime.fromisoformat(data["datetime"])
                self.timestamp = int(datetime.timestamp(_datetime))

        if type(data) is SimpleNamespace:
            self.now = data.now
            self.timestamp = data.timestamp


import requests
import csv
from io import StringIO

class ODKClient:
    def __init__(self):
        self.base = "" #e.g. https://odk-central-testing.swisstph.ch/v1/projects
        self.project = "" #eg, 47
        self.token = "" #eg, base64 encoded from user:pass string
        self.form = "" # e.g. form-1
        self.file = "" # e,g, form-1-attachement.csv

    def auth(self):
        
        url = self.base + "/" + self.project

        payload = {}
        headers = {
        'Authorization': 'Basic '+self.token
        }

        response = requests.request("GET", url, headers=headers, data=payload)

        print(response.text)

    def submissions_data(self):
        
        url = self.base + "/" + self.project + "/forms" + "/" + self.form + "/submissions.csv"

        payload = {}
        headers = {
        'Authorization': 'Basic '+self.token
        }

        response = requests.request("GET", url, headers=headers, data=payload)

        print(response.text)

    def form_attachment(self):

        url = self.base + "/" + self.project + "/forms" + "/" + self.form + "/attachments" +"/" + self.file

        payload = {}
        headers = {
        'Authorization': 'Basic ' + self.token
        }

        response = requests.request("GET", url, headers=headers, data=payload)

        print(response.text)

        # convert string to file object
        f = StringIO(response.text)

        cr = csv.DictReader(f)

        emails = []

        for row in cr:
            emails.append(row["label"])

        print(emails)

        output = {'registered_emails': emails}

        #print(response.text)



from types import SimpleNamespace
class Fields:
    email: str
    data: []

    def __init__(self, data):

        if type(data) is dict:           
            self.email = data["email"]
            if "data" in data:
                self.data = data["data"]
            else:
                self.data = []

        if type(data) is SimpleNamespace:
            self.email = data.email
            self.data  = data.data


class JobMeta:
    hash: str
    created: int
    scheduled: int
    recipients: int
    state: int
    note: str

    def __init__(self, mailjob) -> None:
        
        self.hash = mailjob.hash
        self.created = mailjob.created
        self.scheduled = mailjob.schedule.timestamp
        self.recipients = len(mailjob.recipients)
        self.state = 0
        if mailjob.message.source == 'stdin':
            self.note = "stdin"
        else: 
            self.note = mailjob.message.location



from crontab import CronTab
import shutil

cron = CronTab(user=True)
CRON_COMMENT = 'odk-mailer-cron'

def setup():

    job = find()

    if job and job.is_valid():
        #print("Found valid Cron Job.")
        pass
    else:
        clear()
        create()

def find():

    jobExists = False

    genCron = cron.find_comment(CRON_COMMENT)
    for job in genCron:
        jobExists = True

    if jobExists:
        return job

    return None
        
def create():    

    WHICH_ODK = shutil.which("odk-mailer")
    CRON_COMMAND = WHICH_ODK + ' jobs evaluate --force' # 2>&1 | logger -t mycmd  https://askubuntu.com/a/967798

    job = cron.new(
        #command= '/home/tertek/.cache/pypoetry/virtualenvs/odk-mailer-Dxt_EVX8-py3.10/bin/odk-mailer evaluate --force',
        command = CRON_COMMAND,
        comment= CRON_COMMENT
        )
    # Set to every 15 Minutes
    job.minute.every(1)
    
    cron.write()
    #print("Created Cron Job.")

def clear():
    cron.remove_all(comment=CRON_COMMENT)
    #print("Removed Cron Job.")

def enable():
    job = find()
    if job:
        job.enable()

def disable():
    job = find()
    if job:
        job.enable(False)


from odk_mailer.lib import parser, prompts, utils, globals, smtp, db, log, cron
from odk_mailer.classes.Config import Config
from odk_mailer.classes.Source import Source
from odk_mailer.classes.Fields import Fields
from odk_mailer.classes.Message import Message
from odk_mailer.classes.Schedule import Schedule
from odk_mailer.classes.Mailjob import Mailjob
from odk_mailer.classes.Mailer import Mailer
from email.message import EmailMessage
import os
import sys

def create(input_source, input_fields, input_message, input_schedule):
   
    # source
    if not input_source:
        source_data = prompts.prompt_source()
    else: 
        source_data = parser.parse_source(input_source)
    source = Source(source_data)

    # fields
    headers = source.get_headers()
    if not input_fields:
        fields_data = prompts.prompt_fields(headers)
    else: 
        fields_data = parser.parse_fields(input_fields, headers)
    fields = Fields(fields_data)

    # message
    if not input_message:
        message_data = prompts.prompt_message()
    else:
        message_data = parser.parse_message(input_message)
    message = Message(message_data)

    # schedule
    if not input_schedule:
        schedule_data = prompts.prompt_schedule()
    else:
        schedule_data = parser.parse_schedule(input_schedule)
    schedule = Schedule(schedule_data)        

    # mailjob
    mailjob = Mailjob()
    created = mailjob.create(source, fields, message, schedule)
    
    print(f"Created {created}")
    log.write(f"Created {created}")
    
    ###
    # Features
    ###

    # 1. Reminders 
    # tbd: Reminders
    # reminders have two attributes:
    # total_amount, frequency, e.g. 3 times in total, every hour|day|week|custom frequency
    # after first scheduled send

    # reminders will be stored inside .odk-mailer/reminder/hash_[instance_id].json
    # having updated recipients from non-respondents (calculated from base recipients and respondents) 
    # on a per reminder case

    # https://github.com/tertek/zapier-odk-scripts/blob/main/get_non_responding.py
    # reminders require api connection and following inputs
    # form_register = "test_form_register" #name of form that is used for registration, given as api or csv
    # form_follow = "test_form_follow"  # name of form that is used for follow up
    # # if use_form_attachment
    # form_follow_attached = "test_form_follow_attached" # name of form that is used for follow up; 
    # form_attachment = "follow.csv" # name of form attachment attached to form_follow; 
    # # field config
    # field_email_register = "email_register" # name of email field for registration form, given
    # field_email_follow = "email_follow"
    # field_email_follow_attach = "email"

    # 2. Cron Jobs
    # add cron-job via https://pypi.org/project/python-crontab/ and https://stackabuse.com/scheduling-jobs-with-python-crontab/
    # We will can use a single job, that checks every hour if we have pending Mailjobs, running 'odk-mailer evaluate
    # process reminder tasks as follows: 
    # 1. Perform API request to calculate reminder recipients
    # 2. Send Emails based on calculated recipients
    # 3. Summarize progress

    # 3. Recipients Validation
    # Validate recipients and ignore (e.g. filter) invalid emails to proceed
    # code example
    # if not recipients.validate(email_field):
    #     utils.render_table(["id", "email_field", "error"], recipients.invalidEmails)
    #     ignore_invalid_emails = typer.confirm("Invalid emails found. Would you like to continue although you have invalid emails?")
    
    #     if not ignore_invalid_emails:
    #         raise typer.Exit("\nAborted.")

    # with typer.progressbar(self.data) as progress:
    #     for row in progress:
    #         total += 1
    #         email = row[email_field]
    #         try:
    #             if not email:
    #                 raise EmailNotValidError("Email address missing. Check for missing delimiters (',') in your CSV file.")
    #             # Disable DNS checks since this can be blocked for unknown reasons within network.
    #             validate_email(email, check_deliverability=True)
    #         except EmailNotValidError as e:
    #             invalid = [str(total) ,email, str(e)]
    #             invalid_emails.append(invalid)

def run(hash_or_id, dry=False, verbose=False)->bool:

    mailjob = Mailjob(hash_or_id)
    mailer = Mailer(dry, verbose)

    # check if can be sent
    #
    if db.read_meta(hash_or_id)["state"] != 0:
        utils.abort("Mailjob was already processed. Cannot be run again.")    
    #
    if mailjob.schedule.timestamp> utils.now():
        utils.abort("Mailjob is scheduled for future.")


    hasErrors, errors = mailer.send(mailjob)

    if not dry:
        if not hasErrors:
            mailjob.updateState(1)
            print(f"Success: Run {mailjob.hash}")
            log.write(f"Success: Run {mailjob.hash}")
            return False

        else: 
            mailjob.updateState(2)
            print(f"Failure: Running {mailjob.hash}")
            log.write(f"Failure: Running {mailjob.hash}", "error")
            return False

            # errorsString = ""
            # for idx, error in enumerate(errors):
            #         errorsString = errorsString + idx + ": " +vars(error) + "\n\n"

            # log.write(f"Errors: \n\n{errors}", "error")
            # if verbose:
            #     print(f"Errors:  \n\n{ ','.join(errors) }")


def delete(hash_or_id):

    if not hash_or_id:
        utils.abort("ID is required")
        
    found = db.read_meta(hash_or_id)

    if not found:
        utils.abort("Job not found.")
   
    # tbd: confirm
    
    # deletion from /job/<hash>.json
    path_job = os.path.join(globals.odk_mailer_jobs, found['hash']+'.json')
    if os.path.exists(path_job):
        os.remove(path_job)

    # deletion from meta.json
    jobs_meta_new = list(filter(lambda x: x['hash']!=found['hash'], db.read_meta()))
    db.write_meta(jobs_meta_new)
    
    print("Deleted " + found['hash'])    

def list_jobs():

    jobsMeta = db.read_meta()
    if len(jobsMeta) > 0:
        utils.print_jobs(jobsMeta)

    else: print("There are no mailjobs yet. Create one to see it listed.")

def evaluate(force=False, dry=False, verbose=False):

    jobs = db.read_meta()

    evaluated = list(filter(lambda job: job["state"] == 0 and job["scheduled"] <= utils.now(), jobs))
    total = len(evaluated)

    isCronProcess = sys.stdout.isatty() == False

    if not dry:
        # https://stackoverflow.com/a/2087031/3127170
        if isCronProcess:
            log.write("Evaluate run from cron")
        else: 
            log.write("Evaluate run manually")

        log.write(f"Run evaluate found {total} mailjobs.")

    # # advanced evaluation: addtionally check if job has reminder times that are smaller/equal to now
    # # if true, add to selected list
    # print("Addiitonally checking if we have any valid reminders")
    # for reminder in job["reminders"]:
    #     if reminder["timestamp"] <= utils.now():
    #         evals.append(job["hash"])
    #         break

    if total > 0:
        print(f"Evaluated {total} mailjobs.")

        if force:
            confirmed = True
        else:
            confirmed = prompts.prompt_confirmRun()["confirmRun"]

        if confirmed:
            print(f"Confirmed to run {total} mailjobs.")
            if not dry:
                utils.print_success(f"Confirmed to run {total} mailjobs.")
            for idx, eval in enumerate(evaluated):
                print()
                print(f"Running mailjob #{idx+1} of {total} ({eval['hash']})")
                success = run(eval["hash"], dry, verbose)
                print()
                if not dry and success:
                    utils.print_success(f"Evaluated and run {eval['hash']}")
                    log.write(f"Evaluated and run {eval['hash']}")
    else:
        print("No mailjobs available to be run.")

def config():
    config = Config()
    config.print()

def test(sender, recipient, host, port, username="", password=""):

    smtp_config = [host, port, username, password]

    subject = "Test Mail"
    message = "This is a test mail. Time: " + str(utils.now())

    email = EmailMessage()
    email['From'] = sender
    email['To'] = recipient
    email['Subject'] = subject
    email.set_content(message, subtype="plain")


    print()
    print(f"Sending test mail from {sender} to: {recipient} via: {host}:{port}")
    if username and password:
        print(f"Authentication will be used.")
    print()

    log.write("Sending Test Mail")
    smtp.send(email, True, smtp_config)

def logs():
    log.read()

def crontab_setup():
    cron.setup()

def crontab_enable():
    cron.enable()

def crontab_disable():
    cron.disable()


import typer
from typing import Optional
from typing_extensions import Annotated
from odk_mailer import commands, before
from odk_mailer.classes.Config import Config
import importlib.metadata

app = typer.Typer(
    add_completion=False, 
    pretty_exceptions_enable=True,
    invoke_without_command=True,
    no_args_is_help=True
)
jobs_app= typer.Typer(
    invoke_without_command=True,
    no_args_is_help=True
)
app.add_typer(jobs_app, name="jobs", help="Manage mail jobs")

settings_app= typer.Typer(
    invoke_without_command=True,
    no_args_is_help=True    
)
app.add_typer(settings_app, name="settings", help="Manage settings")

crontab_app = typer.Typer(
    invoke_without_command=True,
    no_args_is_help=True
)
settings_app.add_typer(crontab_app, name="crontabs", help="Manage crontabs")


@app.callback()
def callback(
    version: Annotated[
        Optional[bool], typer.Option("--version", is_eager=True, help="Show version information and exit")
    ] = False,
    config: Annotated[
        Optional[bool], typer.Option("--config", help="Show config information and exit")
    ] = False,
    logs: Annotated[
        Optional[bool], typer.Option("--logs", help="Show live log")
    ] = False,
    ):
    """
    ODK Mailer

    Setup mail jobs by fetching recipients from CSV files or ODK API.

    Run mail jobs immediately or schedule them to be run over time.

    Evaluate available Mailjobs.

    """
    if version: 
        version = importlib.metadata.version('odk_mailer')
        print(version)
        typer.Exit()

    if config:
        config = Config()
        config.print()
        typer.Exit()

    if logs:
        commands.logs()
        typer.exit()


    before.init()

@jobs_app.command("create")
def create(
    source: Annotated[Optional[str], typer.Option("--source", "-s", help="Define source as [type:path|url|api]::[location:str]")]= "",
    fields: Annotated[Optional[str], typer.Option("--fields", "-f", help="Define fields as [email:str]::[data: 'field_1,field_2']")]= "",
    message: Annotated[Optional[str], typer.Option("--message", "-m", help="Define message as [sender:str]::[subject:str]::[source:'stdin'|'path'|'url']::[location:str]::[type:'plain'|'html']")]= "",
    schedule: Annotated[Optional[str], typer.Option("--schedule", help="Define schedule as 'now' or [time:str] in 'YYYY-MM-DD HH:mm' format")]= "",
):
    """
    Create mail job
    """
    commands.create(source, fields, message, schedule)

@jobs_app.command("run")
def run(
    hash: Annotated[str, typer.Argument(help="Hash or id of mailjob to be run")],
    dry: Annotated[bool, typer.Option("--dry", help="Dry run - without sending mails.")] = False,
    verbose: Annotated[bool, typer.Option("--verbose", help="Print out smtp debugging information")] = False,
):
    """
    Run mail job
    """
    commands.run(hash, dry, verbose)

@jobs_app.command("delete")
def delete(
    id: Annotated[str, typer.Argument(help="Hexadecimal hash")]
):
    """
    Delete mail job
    """
    commands.delete(id)

@jobs_app.command()
def list():
    """
    List mail jobs
    """    
    commands.list_jobs()

@jobs_app.command()
def evaluate(
    force: Annotated[bool, typer.Option("--force", help="Enforce to run evaluated mailjobs")] = False,
    dry: Annotated[bool, typer.Option("--dry", help="Dry run without sending mails")] = False,
    verbose: Annotated[bool, typer.Option("--verbose", help="Print out additional SMTP information")] = False
):
    """
    Evaluate mail jobs
    """
    commands.evaluate(force, dry, verbose)

@app.command()
def test(
    sender: Annotated[str, typer.Option("--sender")] = "ODK-Mailer <odk-mailer@freesmtpservers.com>",
    recipient: Annotated[str, typer.Option("--recipient")] = "foo@bar.com",
    host: Annotated[str, typer.Option("--host")] = "smtp.freesmtpservers.com",
    port: Annotated[str, typer.Option("--port")] = "25",
    username: Annotated[str, typer.Option("--user")] = "",
    password: Annotated[str, typer.Option("--pass")] = ""
):
    """
    Send test mail
    """
    commands.test(sender, recipient, host, port, username, password)

@crontab_app.command("setup")
def crontab_setup():
    """
    Setup crontab for odk-mailer
    """
    commands.crontab_setup()


@crontab_app.command("enable")
def crontab_enable():
    """
    Enable crontab for odk-mailer
    """
    commands.crontab_enable()

@crontab_app.command("disable")
def crontab_disable():
    """
    Enable crontab for odk-mailer
    """
    commands.crontab_disable()


import inquirer
from odk_mailer.lib import validators

def prompt_source():
    questions = [
        inquirer.List('type', 
                    message="Select source type:",
                    choices= [
                        ("File from path", "path"),
                        ("File from url", "url"),
                        ("ODK API", "api")
                    ],                      
                    carousel=True
        ),
        inquirer.Path('location',
                    message="Input local path to CSV file",
                    path_type=inquirer.Path.FILE,
                    exists=True,
                    normalize_to_absolute_path=True,
                    ignore=lambda x: x["type"] != "path"
        ),
        inquirer.Text('location',
                    message="Input URL to CSV file",
                    ignore=lambda x: x["type"] != "url",
                    validate=validators.is_url
        ),        
        inquirer.Text('location',
                    message="Input form/attachment name",                      
                    ignore=lambda x: x["type"] != "api"
                    #tbd: add regex valdation
        ),
        inquirer.Text('api_proj',
                    message="Input project id",
                    validate= validators.int_only,
                    ignore=lambda x: x["type"] != "api"
        ),
        inquirer.Text('api_host',
                    message="Input ODK host url",
                    ignore=lambda x: x["type"] != "api",
                    # tbd validate url, https
        ),
        inquirer.Text('api_user',
                    message="Input ODK username",
                    ignore=lambda x: x["type"] != "api"
        ),
        inquirer.Password('api_pass',
                    message="Input ODK password",
                    ignore=lambda x: x["type"] != "api"
        )
    ]
    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def prompt_fields(headers=[]):
    questions = [
        inquirer.List('email',
                    message="Select email field",
                    choices=headers,
                    default='email' if 'email' in headers else "",
                    carousel=True
        ),
        inquirer.Checkbox("data", 
                    message="Select data field(s)",
                    choices= lambda x: filter(lambda y: y != x["email"], headers)
        )        
    ]
    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def prompt_message():
    questions = [
        inquirer.Text('sender',
                    message="Input message sender",
                    default="odk-mailer@freesmtpservers.com",
                    validate=validators.email_address                      
        ),
        inquirer.Text('subject',
                    message="Input message subject",
                    default="ODK-Mailer: New Email"                    
        ),
        inquirer.List('source', 
                    message="Select message source:",
                    choices= [
                        ("Text from stdin", "stdin"),
                        ("File from path", "path"),
                        ("File from url", "url")
                    ],                      
                    carousel=True
        ),        
        inquirer.Text('location',
                    message="Input string",
                    ignore=lambda x: x["source"] != "stdin",
                    validate= validators.not_empty
        ),     
        inquirer.Path('location',
                    message="Input file path",
                    path_type=inquirer.Path.FILE,
                    exists=True,
                    normalize_to_absolute_path=True,
                    ignore=lambda x: x["source"] != "path"
        ),
        inquirer.Text('location',
                    message="Input file url",
                    ignore=lambda x: x["source"] != "url",
                    validate= validators.is_url
        ),        
        inquirer.List('type', 
                    message="Select message type:",
                    choices= [
                        ("Plain", "plain"),
                        ("HTML", "html")
                    ],                      
                    carousel=True
        ),
    ]
    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def prompt_schedule():
    questions = [
        inquirer.List('now',
                    message="Send now or schedule a date and time?",
                    choices=[
                        ("Send now",True),
                        ("Schedule", False)
                    ]
        ),
        inquirer.Text('datetime',
                    message="Input date and time",
                    validate= validators.date_format,
                    ignore=lambda x: x["now"] == True

        )
    ]

    return inquirer.prompt(questions, raise_keyboard_interrupt=True)

def prompt_confirmRun():
    questions = [
        inquirer.Confirm('confirmRun',
                         message="Would you like to run evaluated mailjobs?"
        )
    ]

    return inquirer.prompt(questions, raise_keyboard_interrupt=True)


import csv
from types import SimpleNamespace

class Source:
    type: str
    location: str
    project: int    
    hostname: str
    username: str
    password: str

    def __init__(self, data):

        if type(data) is dict:
            self.type = data["type"]

            if self.type == "path":
                self.location = data["location"]
            
            if self.type == "url":
                self.location = data["location"]

            if self.type == "api":
                self.location = data["location"]
                self.project = data["api_proj"]
                self.hostname = data["api_host"]
                self.username = data["api_user"]
                self.password = data["api_pass"]

        if type(data) is SimpleNamespace:
            self.type = data.type
            self.location = data.location

    def get_headers(self):
        if self.type == 'path':        
            with open(self.location, newline='') as f:
                dialect = csv.Sniffer().sniff(f.read(1024))
                f.seek(0)
                reader = csv.DictReader(f, skipinitialspace=True, delimiter=dialect.delimiter)
                headers = reader.fieldnames
            return headers
        
        raise Exception("Source type is not yet implemented" )

    def get_rows(self):
        if self.type == 'path':        
            with open(self.location, newline='') as f:
                dialect = csv.Sniffer().sniff(f.read(1024))
                f.seek(0)
                reader = csv.DictReader(f, skipinitialspace=True, delimiter=dialect.delimiter)
                rows = []
                for row in reader:
                    rows.append(row)

            return rows
        
        raise Exception("Source type is not yet implemented" )
       


from inquirer import errors
import re

def int_only(_, current):
    re_integer = "^[0-9]*$"
    if not re.search(re_integer, current):
        raise errors.ValidationError('', 'Invalid project id format. Has to be a number')
    return True

def date_format(_, current):
    # tbd: must be future
    re_YYYY_MM_DD_hh_mm = "[0-9]{4}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1]) (2[0-3]|[01][0-9]):[0-5][0-9]"
    if not re.search(re_YYYY_MM_DD_hh_mm, current):
        raise errors.ValidationError('', 'Invalid date format. Has to be YYYY-MM-DD hh:mm')       
    return True

def email_address(_, current):
    re_email_address  = "^\S+@\S+\.\S+$"
    if not re.search(re_email_address, current):
        raise errors.ValidationError('', 'Invalid email address format.')
    return True

def not_empty(_, current):
    if current == "":
         raise errors.ValidationError('', 'Content cannot be empty.')
    return True

def is_url(_, current):
    re_url = "((http|https)\:\/\/)?[a-zA-Z0-9\.\/\?\:@\-_=#]+\.([a-zA-Z]){2,6}([a-zA-Z0-9\.\&\/\?\:@\-_=#])*"
    # https://stackoverflow.com/a/48689681/3127170
    if not re.search(re_url, current):
        raise errors.ValidationError('', 'Invalid URL format.')
    return True


import logging
from pathlib import Path
import typer
import time
import os

# tbd: redirect smtp debug to log file: https://stackoverflow.com/a/7303587/3127170

FILE_NAME_LOG = 'odk-mailer.log'
APP_NAME = 'odk-mailer'
app_dir=typer.get_app_dir(APP_NAME, force_posix=True)
log_path: Path = Path(app_dir) / FILE_NAME_LOG


logging.basicConfig(
    filename=log_path, 
    encoding='utf-8', 
    level=logging.DEBUG, 
    format='%(asctime)s: %(levelname)s: %(message)s'
)

def write(msg, type="info"):

    if type == "info":
        logging.info(msg)

    if type=="error":
        logging.error(msg)


def read():
    def follow(thefile):
        '''generator function that yields new lines in a file
        '''
        # seek the end of the file
        thefile.seek(0, os.SEEK_END)
        
        # start infinite loop
        while True:
            # read last line of file
            line = thefile.readline()
            if not line:
                time.sleep(0.1)
                continue

            yield line
    
    logfile = open(log_path,"r")
    loglines = follow(logfile)
    
    print(f"Following on logfile: {log_path}..")
    for line in loglines:
        print(line)